--  rat.e
--  jbabor@paradise.net.nz
--  00-11-22
--  version 2.50

without warning
without type_check

include machine.e

global constant
    lbp     = 1,    -- left button pressed
    rbp     = 2,    -- right button pressed
    mbp     = 4,    -- middle button pressed
    rmoved  = 8,    -- rat moved
    rat_events = 15 -- all rat events combined (mask)

constant regs = repeat(0,10)

global integer
    rat_event,      -- rat moved or a rat button is pressed
    rx,             -- current rat horizontal coordinate
    ry,             -- current rat vertical coordinate
    rat_visible     -- boolean, rat pointer visible or hidden (1 or 0)

sequence pointer    -- rle pointer image
integer maxrx, minrx, maxry, minry, hotx, hoty,
    lastx,         -- saved (last drawn) rat horizontal coordinate
    lasty          -- saved (last drawn) vertical coordinate

global function compile_rat_pointer(
    integer x,      -- 'hot' x
    integer y,      -- 'hot' y
    sequence s      -- bitmap, 2-d color pointer image
    )
    --  run length encode pointer image for faster display later

    sequence si,u,ui
    integer h,inrun,count,w,k

    h=length(s)
    u={}
    for i=1 to length(s) do
        inrun = 0
        count = 0
        si = s[i]
        ui = {}
        w = length(s[i])
        for j=1 to w do
            if si[j] = 0 then
                if inrun then
                    ui &= count & si[k..j-1]
                    count = 1
                    inrun = 0
                else
                    count += 1
                end if
            else
                if inrun then
                    count += 1
                else
                    ui &= count
                    k = j
                    count = 1
                    inrun = 1
                end if
            end if
        end for
        if inrun then
            ui &= count & si[k..w]
        end if
        u=append(u,ui)
    end for
    return prepend(u, {x,y})
end function -- compile_rat_pointer

procedure draw_rat()
    --  display rle pointer image

    sequence rlei
    integer j,run,n,u,x,y

    x=lastx-hotx
    y=lasty-hoty
    for i=1 to length(pointer) do
        rlei=pointer[i]
        j=1
        u=x
        run=0
        while j<length(rlei) do
            n=rlei[j]
            if run then
                pixel(xor_bits(get_pixel({u,y,n}),rlei[j+1..j+n]),{u,y})
                j+=n+1
            else
                j+=1
            end if
            run=not run
            u+=n
        end while
        y+=1
    end for
end procedure -- draw_rat

global procedure show_rat()
    if not rat_visible then draw_rat() end if
    rat_visible=1
end procedure -- show_rat

global procedure hide_rat()
    if rat_visible then
        draw_rat()      -- erase
        rat_visible = 0
    end if
end procedure -- hide_rat

global procedure rat()
    sequence r
    integer m,
    rat_moved       -- boolean, if true then the rat pointer
                    -- has moved since the last rat() call

    r = regs
    r[REG_AX] = #B
    r = dos_interrupt(#33,r)  -- relative motion counters

    m=r[REG_CX]
    if m then
        rx += m - 2 * and_bits(m, #8000)
        if rx < minrx then rx = minrx
        elsif rx > maxrx then rx = maxrx
        end if
    end if

    m = r[REG_DX]
    if m then
        ry += m - 2 * and_bits(m, #8000)
        if ry<minry then ry = minry
        elsif ry>maxry then ry = maxry
        end if
    end if

    r = regs
    r[REG_AX] = 3
    r = dos_interrupt(#33,r)  -- poll rat
    rat_event = r[REG_BX]

    rat_moved = lastx != rx or lasty != ry
    if rat_moved then
        rat_event += rmoved
        if rat_visible then
            draw_rat()      -- erase image at previous location
            lastx = rx
            lasty = ry
            draw_rat()      -- show it again
        else
            lastx = rx
            lasty = ry
        end if
    end if
end procedure -- rat

global procedure confine_rat(
    integer x1, integer y1,     -- left top corner of screen cage
    integer x2, integer y2      -- right bottom corner
    )
    minrx = x1
    minry = y1
    maxrx = x2
    maxry = y2
    if rx<minrx then
        rx=minrx
    elsif rx>maxrx then
        rx=maxrx
    end if
    if ry<minry then
        ry=minry
    elsif ry>maxry then
        ry=maxry
    end if
end procedure -- confine_rat

global procedure set_rat_xy(integer x, integer y)
    -- places the rat pointer where you want it

    if x<minrx then
        x=minrx
    elsif x>maxrx then
        x=maxrx
    end if
    if y<minry then
        y=minry
    elsif y>maxry then
        y=maxry
    end if
    rx=x
    ry=y
    rat()
end procedure  -- set_rat_xy

global procedure rat_pointer(
    sequence p  -- rle pointer image with prepended hot spot coordinates
    )
    --  select new pointer

    integer flag

    flag = rat_visible      -- save current rat visibility
    hide_rat()              -- make rat invisible
    hotx = p[1][1]-1
    hoty = p[1][2]-1
    pointer = p[2..length(p)]
    if flag then show_rat() end if  -- make new pointer visible
end procedure -- rat_pointer

global procedure init_rat()
    --  initialize - must be called *after* graphics mode is set
    --  set pointer in middle of screen, hidden
    --  confine rat to visible screen area

    sequence vc

    vc = machine_func(13,0)           -- video_config()
    rat_event = 0
    rx = floor(vc[5]/2)
    lastx = rx
    ry = floor(vc[6]/2)
    lasty = ry
    minrx = 0
    maxrx = vc[5]-1
    minry = 0
    maxry = vc[6]-1
end procedure -- init_rat

global procedure wait_while (integer pressed_button)
    --  wait while specified rat button remains pressed

    while and_bits(rat_event, pressed_button) do
        rat()      -- keep rat pointer alive
    end while
end procedure -- wait_while

global function rat_zone(sequence zones)
    --  scan zones for rat pointer presence
    --  zones is sequence of rectangular zones to be tracked
    --  each zone is defined by {x1, y1, x2, y2) sequence, where x1, y1
    --  and x2, y2 are top left and  bottom right diagonal points resp.
    --  return index of the *first* zone that includes the rat pointer,
    --  or zero if none found

    sequence zi

    for i=1 to length(zones) do
        zi=zones[i]
        if rx>=zi[1] and
           rx<=zi[3] and
           ry>=zi[2] and
           ry<=zi[4] then
           return i
        end if
    end for
    return 0
end function -- rat_zone


----------------------------------------------------------------------
--  dragbox routines
----------------------------------------------------------------------

sequence dbox   -- drag box state: {x1,y1,x2,y2,color}

procedure safe_xor_box(integer x1, integer y1, integer x2, integer y2,
    integer color)
    --  draw line box xor-ed with background
    --  x1,y1 & x2,y2 are diagonal points
    --  safe, because order of points is immaterial

    sequence s
    integer dummy, len

    if x1>x2 then   -- safety first
        dummy=x2
        x2=x1
        x1=dummy
    end if
    if y1>y2 then
        dummy=y2
        y2=y1
        y1=dummy
    end if
    len=x2-x1+1
    s=repeat(color,len)
    pixel(xor_bits(get_pixel({x1,y1,len}),s),{x1,y1})
    pixel(xor_bits(get_pixel({x1,y2,len}),s),{x1,y2})
    for y=y1+1 to y2-1 do
        pixel(xor_bits(get_pixel({x1,y}),color),{x1,y})
        pixel(xor_bits(get_pixel({x2,y}),color),{x2,y})
    end for
end procedure -- safe_xor_box

global procedure erase_drag_box()
    --  erase drag box if one is drawn (visible)

    integer visible     -- rat visibility flag

    if dbox[5] then     -- box visible
        visible=rat_visible
        hide_rat()
        safe_xor_box(dbox[1],dbox[2],dbox[3],dbox[4],dbox[5])
        dbox[5]=0       -- box invisible
        if visible then show_rat() end if   -- restore rat
    end if
end procedure -- erase_drag_box

global function drag_box(
    integer color,              -- drag box color
    integer button_pressed
    )
    -- return drag box data: {x1,y1,x2,y2,color}
    -- this function integrates drag box with rat xor pointer:
    -- continually updates drag box data while specified rat
    -- button remains pressed
    -- color is drag box color (true only on BLACK background)

    integer x1, y1, x2, y2, visible

    visible = rat_visible

    if and_bits(button_pressed, rat_event) then
        erase_drag_box()
        x1=rx
        y1=ry
        x2=x1
        y2=y1
        safe_xor_box(x1,y1,x2,y2,color)
        while and_bits(rat_event, button_pressed) do
            if rx!=x2 or ry!=y2 then
                hide_rat()
                safe_xor_box(x1,y1,x2,y2,color)
                x2=rx
                y2=ry
                safe_xor_box(x1,y1,x2,y2,color)
                if visible then
                    show_rat()
                end if
            end if
            rat()       -- keep rodent alive
        end while
        dbox = {x1,y1,x2,y2,color}
    end if
    return dbox
end function -- drag_box

-- init drag box data
dbox={0,0,0,0,0}    -- {x1,y1,x2,y2,color}
                    -- box not drawn if color=0


----------------------------------------------------------------------
--  various rat pointers
----------------------------------------------------------------------

global constant RAT_ARROW = compile_rat_pointer(1, 1,{
        "/",
        "//",
        "/'/",
        "/''/",
        "/'''/",
        "/''''/",
        "/'''''/",
        "/''''''/",
        "/'''''''/",
        "/'''''////",
        "/'''''/",
        "/'//''/",
        "//  /''/",
        "/   /''/",
        "     /''/",
        "     ////"}-32)

global constant  RAT_TRIANGLE = compile_rat_pointer(1,1,{
        "///////////",
        "/'''''''''",
        "/''''''''",
        "/'''''''",
        "/''''''",
        "/'''''",
        "/''''",
        "/'''",
        "/''",
        "/'",
        "/"}-32)

global constant RAT_CROSS = compile_rat_pointer(7,7,{
        "      /",
        "      /",
        "      /",
        "      /",
        "      /",
        "",
        "/////   /////",
        "",
        "      /",
        "      /",
        "      /",
        "      /",
        "      /"}-32)

global constant RAT_IBEAM = compile_rat_pointer(3,7,{
        "// //",
        "  /",
        "  /",
        "  /",
        "  /",
        "  /",
        "  /",
        "  /",
        "  /",
        "  /",
        "  /",
        "  /",
        "// //"}-32)

global constant RAT_LTRB = compile_rat_pointer(6,6,{
        "/////",
        "/''/",
        "/'''/",
        "//'''/",
        "/ /'''/",
        "   /'''/",
        "    /'''/ /",
        "     /'''//",
        "      /'''/",
        "       /''/",
        "      /////"}-32)

global constant RAT_RTLB = compile_rat_pointer(6,6,{
        "      /////",
        "       /''/",
        "      /'''/",
        "     /'''//",
        "    /'''/ /",
        "   /'''/",
        "/ /'''/",
        "//'''/",
        "/'''/",
        "/''/",
        "/////"}-32)

global constant RAT_LR = compile_rat_pointer(8,4,{
        "   /       /",
        "  //       //",
        " /'/////////'/",
        "/'''''''''''''/",
        " /'/////////'/",
        "  //       //",
        "   /       /"}-32)

global constant RAT_TB = compile_rat_pointer(4,8,{
        "   /",
        "  /'/",
        " /'''/",
        "///'///",
        "  /'/",
        "  /'/",
        "  /'/",
        "  /'/",
        "  /'/",
        "  /'/",
        "  /'/",
        "///'///",
        " /'''/",
        "  /'/",
        "   /"}-32)

global constant RAT_PENCIL = compile_rat_pointer(1,1,{
        ",,///",
        ",///'/",
        "///'/'/",
        "//'''/'/",
        "/'/'''/'/",
        " /'/'''/'/",
        "  /'/'''/'/",
        "   /'/'''/'/",
        "    /'/'''/'/",
        "     /'/'''/",
        "      /'/'/",
        "       /'/",
        "        /"}-32)

global constant RAT_HGLASS = compile_rat_pointer(5,7,{
        "/////////",
        "/       /",
        "/       /",
        " /'''''/",
        "  /'''/",
        "   /'/",
        "    /",
        "   / /",
        "  /   /",
        " /  '  /",
        "/ ''''' /",
        "/'''''''/",
        "/////////"}-32)

global constant RAT_NOGO = compile_rat_pointer(7,7,{
        "     ////     ",
        "   //$$$$//   ",
        "  /$$$//$$$/  ",
        " /$$//  /$$$/ ",
        " /$/   /$$/$/ ",
        "/$$/  /$$//$$/",
        "/$/  /$$/  /$/",
        "/$$//$$/  /$$/",
        " /$/$$/   /$/ ",
        " /$$$/  //$$/ ",
        "  /$$$//$$$/  ",
        "   //$$$$//   ",
        "     ////     "}-32)

global constant RAT_RESIZE = compile_rat_pointer(9,9,{
        "        /        ",
        "       /'/       ",
        "      /'''/      ",
        "     /'''''/     ",
        "     ///'///     ",
        "   //  /'/  //   ",
        "  /'/  /'/  /'/  ",
        " /''////'////''/ ",
        "/'''''''''''''''/",
        " /''////'////''/ ",
        "  /'/  /'/  /'/  ",
        "   //  /'/  //   ",
        "     ///'///     ",
        "     /'''''/     ",
        "      /'''/      ",
        "       /'/       ",
        "        /        "}-32)

global constant RAT_HAND = compile_rat_pointer(3,1,{
        "  ///",
        "  / /",
        "  / /",
        "  / /",
        "  / /",
        "  / /",
        "  / ///////",
        "  / / / / /",
        "/// / / / /",
        "/ / / / / /",
        "/ /       /",
        "/ /       /",
        "/         /",
        "/         /",
        "/         /",
        "///////////"}-32)

global constant RAT_HAND1 = compile_rat_pointer(3,1,{
        "  ///",
        "  /'/",
        "  /'/",
        "  /'/",
        "  /'/",
        "  /'///",
        "  /'/'///",
        "  /'/'/'///",
        "///'/'/'/'/",
        "/'/'/'/'/'/",
        "/'/'''''''/",
        "/'/'''''''/",
        "/'''''''''/",
        "/'''''''''/",
        "/'''''''''/",
        "///////////"}-32)


-- init rat ----------------------------------------------------------
rat_visible = 0         -- let's make rat initially hidden
rat_pointer(RAT_TRIANGLE)
--rat_pointer(RAT_ARROW)

