--  tsp.ex : travelling salesman problem
--  jbabor@paradise.net.nz
--  00-11-22
--  version 3.00

without warning
without type_check

include button.e
include checkbox.e
include rbuttons.e
include isel.e
include zone.e
include reader.e

include modern28.e          -- my favourite large font
include vga8x16.e           -- fixed pitch font

constant pi = 3.14159
constant x0 = 20            -- grid left edge
constant y0 = 20            -- grid top edge
constant R = 220
constant RAND = 1, POLY = 2, GRID = 3, SCI = 4
constant m = 4.5            -- Sci Am display multiplier
constant mu = 1/m           -- Sci Am path scaler
constant e = 0.00001        -- rounding error limit
object o
sequence b,best_path,buffer,l,node,path,rb,z,user_path
atom best,current,user_total,previous,worst,t,dt
integer end_of_buffer,iteration,previous_iteration,maxi,maxsize,maxstep,
    n,ni,buffer_size,zo,re,worked,rbs,kind,et,old_et

constant hs = {             -- help lines
    "",
    -24 & "~What Is This?~",
    "",
    -16 & "Just another lousy solution of the '~Travelling Salesman Problem~'...",
    "",
    "The goal is to find the ~shortest~ route for a given set of cities. Each city",
    "must be visited ~once~, and the route has to be a ~round~ trip.",
    "",
    "",
    -24 & "~How?~",
    "",
    -16 & "Just press the buttons! Seriously... ;-)",
    "",
    "The initial node count of ~36~ can be reset to anything from ~6~ to ~200~ nodes.",
    "Note, it is often faster and certainly much easier to use the '~hot~' keys, ~left~",
    "and ~right~ arrows in this case, than the mouse. ~PageUp~ and ~PageDn~ keys",
    "are also quite useful within this ~help~ window to avoid excessive repaints.",
    "",
    "The '~new~' button generates a new set of nodes depending on the current",
    "selection within the radio button cluster and the chosen number of nodes.",
    "",
    "The '~random~' button gives a new set of randomly dispersed nodes.",
    "",
    "The '~polygon~' button is for testing only: it produces a regular polygon with",
    "the selected number of vertices and a random path through them.",
    "",
    "The '~grid~' button is also for testing only. It generates the first ~n~ nodes in",
    "a ~square~ matrix, once again randomly connected.",
    "",
    "The '~sci am~' button will use the first ~n~ nodes, randomly shuffled, from the",
    "famous 200 point suite published in the ~Scientific American~ magazine.",
    "",
    "Every time you press the '~step~' button, the program searches for a better",
    "solution than the current one, and when it is found it is displayed and the",
    "program simply stops. It also stops when an arbitrary cut-off of ~n*n*n~",
    "iterations is reached (where ~n~ is the number of nodes, cities).",
    "",
    "The '~run~' button will take you right through to this cut-off in one go.",
    "",
    "The program automatically doubles the cut-off each time you attempt to",
    "step or run beyond this limit. But if you keep pressing the buttons long",
    "enough, the program will eventually run out of buffered alternative",
    "solutions and it will stops completely. The '~run~' and the '~step~' button will",
    "become grayed, inactive. This, unfortunately, can some times happen,",
    "especially with larget sets, even before an optimal solution is found.",
    "",
    "The '~save~' button becomes active only when '~sci am~' mode is selected, and",
    "the node count is exactly 200. It allows you to save the total path, iteration",
    "count and the node order sequence into ~tsp.sav~ file in human readable",
    "form...",
    "",
    "",-24 & "~Can You Beat IT ?~",
    "",
    -16 & "You can try. Click on a node and the background will turn ~red~, indicating",
    "you are in the input mode. The last clicked node is always shown in "&#10E&"yellow"&#100,
    "color. The accumulated path length is displayed in the small red panel on",
    "the right hand side. If you make a mistake, you can start again by simply",
    "pressing the '~redo~' button.",
    "",
    "",
    -24 & "~The Digital Display.~",
    "",
    -16 & "The display hardly requires any explanations. The little black window is a",
    "simple ~elapsed time~ clock. Only two best solutions are displayed in the grey",
    "windows. The first column contains the total paths in pixel units. These",
    "values are suitably scaled for the set from the ~Scientific American~.",
    "Corresponding iterations appear in the second column.",
    "",
    "",
    -24 & "~The Method.~",
    "",
    -16 & "Very simple and systematic: the program starts with a completely random",
    "path through all ~n~ nodes. Then it tries to improve the path by ~reversing~ a",
    "fixed length segment of it starting at every node around it. If the new path",
    "is one of the best ~2*n~ paths (arbitrary), it is stored in a buffer. Only the",
    "currently best unexplored path is worked on using progressively shorter",
    "segments, starting with ~floor((n+1)/2)~ nodes and going down to ~one~.",
    "",
    "Enjoy!   jiri"}

procedure show_time(sequence s)
    bar(BLACK,480,300,70,17)
    select_font(vga8x16)
    reset_font()
    set_attributes(NONE)
    set_foreground(12)
    justify(RIGHT)
    setxy(545,301)
    write(s)
end procedure -- show_time

procedure update_timer()
    sequence s
    integer h,m

    dt += time()-t
    t = time()
    et = floor(dt)              -- elapsed time
    if et > old_et then
        if et > 3600 then
            h = floor(et/3600)
            m = floor(et/60) - 60*h
            s = sprintf("%d:%02d:%02d", {h, m, remainder(et, 60)})
        else
            s = sprintf("%0d:%02d", {floor(et/60),remainder(et, 60)})
        end if
        show_time(s)
        old_et = et
    end if
end procedure -- update_timer

procedure show_headers()
    bar(3,480,5,155,235)        -- title panel
    select_font(modern28)
    reset_font()
    justify(CENTRE)
    set_foreground(11)
    set_attributes(SHADOW)
    set_shadow_color(0)
    set_extra_char_pitch(1)
    setxy(558,25)
    write("travelling")
    setxy(558,60)
    write("salesman")
    setxy(558,95)
    write("again ...")
    set_extra_char_pitch(0)

    select_font(default_font)
    set_foreground(7)
    setxy(558,145)
    write("version 3.00")
    setxy(558,165)
    write("november 2000")
    set_foreground(8)
    setxy(558,195)
    write("jiri babor")
    set_foreground(7)
    setxy(558,215)
    write("jbabor@paradise.net.nz")

    set_attributes(NONE)
    justify(RIGHT)
    set_foreground(1)
    setxy(545,248)
    write("total path")
    setxy(630,248)
    write("iteration")

    justify(LEFT)
    set_foreground(7)
    setxy(485,342)
    write("your total:")
end procedure -- show_headers

procedure set_lengths()
    atom dx,dy
    l = repeat(repeat(0,n),n)
    for i=1 to n do
        for j=i+1 to n do
            dx = node[i][1] - node[j][1]
            dy = node[i][2] - node[j][2]
            l[i][j] = sqrt(dx*dx + dy*dy)
            l[j][i] = l[i][j]
        end for
    end for
end procedure -- set_lengths

procedure show_user_total()
    bar(12,559,339,76,17)       -- erase right red total field
    select_font(vga8x16)
    reset_font()
    set_attributes(NONE)
    justify(RIGHT)
    set_foreground(15)
    setxy(630,340)
    if kind=SCI then
        write(sprintf("%.1f", mu*user_total))
    else
        write(sprintf("%.1f", user_total))
    end if
end procedure -- show_user_total

procedure show_previous()
    select_font(vga8x16)
    reset_font()
    set_attributes(NONE)
    set_foreground(15)
    justify(RIGHT)

    bar(3,480,260,70,17)
    setxy(545,261)
    if kind = SCI then
        write(sprintf("%.1f", mu*previous))
    else
        write(sprintf("%.1f", previous))
    end if
    bar(3,553,260,82,17)
    setxy(630,261)
    write(sprint(previous_iteration))
end procedure -- show_previous

procedure show_best()
    select_font(vga8x16)
    reset_font()
    set_attributes(NONE)
    set_foreground(15)
    justify(RIGHT)

    bar(3,480,280,70,17)
    setxy(545,281)
    if kind = SCI then
        write(sprintf("%.1f", mu*best))
    else
        write(sprintf("%.1f", best))
    end if
    bar(3,553,280,82,17)
    setxy(630,281)
    write(sprint(iteration))
end procedure -- show_best

function path_length(sequence s)
    atom len

    len = l[s[1]][s[n]]
    for i=1 to n-1 do
        len += l[s[i]][s[i+1]]
    end for
    return len
end function -- path_length

function shuffle(sequence s)
    -- random sequence shuffle

    object k
    integer j,l

    l = length(s)
    for i=l to 2 by -1 do
        j = rand(i)
        k = s[j]
        s[j] = s[i]
        s[i] = k
    end for
    return s
end function -- shuffle

procedure set_buffers()
    set_lengths()
    best = path_length(path)        -- current path length
    worst = best
    buffer = {{path , maxstep, 1, best}}
    best_path = path
    buffer_size = 1
    worked = 1                      -- currently worked record
end procedure -- set_buffers

procedure random_nodes()
    -- generate new set of randomly spaced nodes

    node = repeat({0,0},n)
    path = repeat(0,n)
    --set_rand(123)                 -- uncomment to seed randomizer
    for i=1 to n do
        path[i] = i
        node[i][1] = rand(460)+9    -- x
        node[i][2] = rand(460)+9    -- y
    end for
end procedure -- random_nodes

procedure poly_nodes()
    -- generate set of nodes spaced as regular polygon

    atom a,da

    node = repeat({0,0},n)
    path = repeat(0,n)

    da = 2*pi/n
    a = 0.0
    for i=1 to n do
        path[i] = i
        node[i][1] = round(R*sin(a))+240    -- x
        node[i][2] = round(R*cos(a))+240    -- y
        a += da
    end for
    path = shuffle(path)
end procedure -- poly_nodes

procedure grid_nodes()
    integer d,m,x,y

    node = repeat({0,0},n)
    path = repeat(0,n)
    m = floor(sqrt(n))
    if m*m < n then m += 1 end if
    d = floor(440/(m-1))

    x = x0
    y = y0
    for i=1 to n do
        path[i] = i
        node[i][1] = x
        node[i][2] = y
        if remainder(i,m) = 0 then
            x = x0
            y += d
        else
            x += d
        end if
    end for
    path = shuffle(path)
end procedure -- grid_nodes

constant sa = {         -- 200 x,y points from Scientific American
51.387066,17.572557,30.863369,53.453171,94.763023,17.172766,70.223091,22.641682,
49.476608,12.469863,8.389538 ,38.962982,27.723015,36.805321,98.345894,53.538620,
76.567886,64.647362,76.714378,78.023621,82.296211,15.192114,62.547684,31.467634,
34.690392,91.720329,51.976074,40.116581,60.676899,78.542435,93.154701,86.993011,
86.654259,67.452011,75.841545,58.189644,38.923306,35.563221,20.023193,82.692955,
41.590626,46.351513,97.918640,12.643819,21.262245,95.846428,73.747978,40.903957,
78.011414,75.789665,95.684685,2.807703 ,31.873531,75.695061,24.298838,58.955658,
4.339732 ,95.605331,31.913206,5.935850 ,44.187752,91.503647,57.225258,11.883907,
56.978058,25.205236,49.586475,23.673208,47.697380,40.607929,87.301247,42.695396,
35.822628,38.196968,4.315317 ,16.058840,52.235481,69.658501,9.707938 ,40.083012,
77.343056,24.481949,34.281441,22.998749,29.786066,30.454420,88.720360,3.665273 ,
65.114288,39.860226,67.629021,73.259682,93.780327,23.328348,83.849602,96.722313,
77.864922,43.150120,67.409286,80.938141,15.875729,27.988525,13.531907,86.419266,
75.020599,20.798365,13.995789,29.459517,80.281990,21.893978,56.309700,71.559799,
19.754631,98.983734,25.003815,43.061615,75.527206,86.092712,89.480270,97.808769,
39.539780,43.220314,12.714011,45.768608,23.783075,98.605301,65.282143,60.423595,
24.188971,45.487839,78.997162,7.879879 ,47.639393,15.259254,24.573503,94.500565,
61.403240,98.818932,47.727898,79.970703,74.419388,38.071842,47.990356,52.690208,
9.808649 ,59.422588,34.717857,14.334544,77.953430,71.098969,44.615009,70.458084,
9.530931 ,96.282845,55.131687,74.025696,57.902767,63.786736,78.167061,18.790247,
30.210272,28.281504,68.401138,29.291666,56.538589,41.843929,30.658895,44.453262,
56.569111,48.792992,60.664692,41.584522,13.040559,25.595875,3.573717 ,97.711113,
11.450544,37.806328,64.671776,35.044403,55.305641,35.840939,56.544693,47.563099,
16.367077,61.522263,17.215490,55.470444,29.221474,87.215797,83.507797,84.490494,
89.553513,59.477524,54.057434,16.818750,65.495773,69.051178,26.383251,10.669271,
81.493576,19.135105,42.329174,35.184788,83.922852,13.733329,26.267281,17.725151,
47.990356,38.016907,50.483719,50.276192,35.190895,52.558975,12.060915,51.954712,
60.713524,73.290199,55.687122,34.412670,80.196541,59.099094,26.691488,67.067474,
55.214088,78.893402,88.772240,89.001129,6.808680 ,80.059204,90.737633,64.412369,
16.513565,30.133976,16.626484,28.516495,84.197517,53.633228,3.634754 ,20.719017,
2.121036 ,35.813473,62.147892,52.003540,54.603718,15.366070,82.335884,3.332621 ,
2.594073 ,37.812431,61.632130,2.038636 ,62.657551,91.521957,37.479782,72.945343,
39.582508,98.226875,59.727776,11.230811,22.159491,79.918823,87.066254,73.824272,
1.361125 ,73.955505,41.834774,36.204109,20.389416,18.314157,7.626575 ,11.554308,
15.912351,78.826256,4.034547 ,79.064301,59.901730,40.260017,22.904140,18.280586,
61.430706,33.188881,60.515152,96.411026,37.806328,18.439283,30.008850,5.417036 ,
14.401685,1.043733 ,88.485367,95.803703,62.590412,95.562607,63.103123,3.912473 ,
35.129856,14.633625,10.602130,19.742424,8.392591 ,2.682577 ,94.573807,91.964478,
90.798668,86.562698,14.893033,17.178869,6.820887 ,65.102081,73.686943,10.242012,
16.000854,9.396649 ,12.170781,2.462844 ,76.232185,95.696892,2.789392 ,64.659569,
10.803553,42.793053,30.973236,1.858577 ,88.531143,75.786613,50.953705,16.577654,
76.293221,88.085571,49.955750,87.514877,73.479416,23.502304,5.160680 ,60.548721,
87.597275,50.401318,67.839592,98.941010,60.478531,49.635303,58.973969,89.550461,
4.461806 ,88.286995,10.821863,52.012695,57.884457,0.991852 ,38.706627,47.712639,
19.293802,50.788902,77.501755,35.438095,69.777519,91.280861,67.094940,70.577103,
42.689293,2.090518 ,21.295816,94.759972,50.282295,19.425032,64.473404,12.796411,
26.493118,33.603931,70.369576,3.811762 ,95.361183,75.475327,87.435532,63.399151,
24.353771,63.579212,85.048981,23.737297,72.096924,33.921322,5.020295 ,48.500015,
89.748833,24.228645,52.751244,49.446091,85.506760,34.589680,12.384411,21.552172,
11.548204,36.320080,20.398571,43.607899,82.830284,50.972015,81.951355,41.071812}

procedure sci_nodes()
    -- use first n nodes sa array (Sci Am)
    node = repeat({0,0},n)
    path = repeat(0,n)
    for i=1 to n do
        path[i] = i
        node[i][1] = 4.5 * sa[2*i-1] + 14
        node[i][2] = 4.5 * sa[2*i] + 14
    end for
    path = shuffle(path)    -- add a bit of variety
end procedure -- sciam_nodes

procedure draw_screen()
    hide_rat()
    bar(7,0,0,hsize,vsize)              -- blank screen
    bar(1,5,5,470,470)                  -- field
    frame({{3,3},{11,11}},5,5,470,470)  -- field frame
    frame({{8},{3}},480,385,155,40)     -- radio buttons
    bar(1,480,320,50,16)
    bar(1,532,320,50,16)
    bar(1,584,320,50,16)
    bar(12,480,339,76,17)               -- left red total fielf
    bar(12,559,339,76,17)               -- right red total field
    show_headers()
    show(b & ni & rbs)
end procedure -- draw_screen

procedure draw_path(integer c)
    sequence s

    s = repeat(0, n)
    for i=1 to n do
        s[i] = node[best_path[i]]
    end for
    polygon(c,0,s)
end procedure -- draw_path

procedure draw_nodes()
    for i=1 to n do
        ellipse(15,1,node[i]-2,node[i]+2)
    end for
end procedure -- draw_nodes

procedure redraw()
    hide_rat()
    bar(1,7,7,466,466)             -- blue field
    draw_path(9)
    draw_nodes()
    show_rat()
end procedure -- redraw

procedure erase_user_entry()
    user_path = {}
    user_total = 0
    bar(12,559,339,76,17)          -- erase right red total field
end procedure

procedure quit()
    exitf = true                    -- exit flag
end procedure -- quit

function identical(integer i)
    sequence p,pr
    integer j,l

    -- find the lowest buffer entry of same path length
    l = length(buffer)
    while i < l do                                  -- go down
        if abs(buffer[i+1][4] - current) < e then
            i += 1
        else
            exit
        end if
    end while

    while i and abs(buffer[i][4] - current) < e do  -- now go up again
        p = buffer[i][1]        -- path
        j = find(path[1],p)
        p = p[j..n] & p[1..j-1]
        pr = reverse(p)
        pr = pr[n] & pr[1..n-1]
        if equal(p, path) or equal(pr, path) then
            return 0
        end if
        i -= 1
    end while
    return i+1
end function -- identical

function bsearch()
    --  special binary search function for buffer inserts
    --  return *zero* if identical record already exists
    --  otherwise return insertion index

    atom a
    integer hi,lo,mi

    lo = 1
    hi = length(buffer)
    while lo<=hi do
        mi = floor((lo+hi)/2)
        a = buffer[mi][4]
        if abs(a - current) < e then    -- near identical
            return identical(mi)
        elsif current < a then
            hi = mi-1
        else
            lo = mi+1
        end if
    end while
    return lo
end function -- bsearch

procedure insert()
    integer i

    i = bsearch()
    if i then
        buffer = buffer[1..i-1] & {{path,maxstep,1,current}} & buffer[i..buffer_size]
        buffer_size = length(buffer)
        if buffer_size > maxsize then
            buffer_size = maxsize
            buffer = buffer[1..buffer_size]
        end if
        worst = buffer[buffer_size][4]
        if i < worked then
            worked = i
        end if
    end if
end procedure -- insert

procedure get_worked()
    while buffer[worked][3] > n do  -- skip searched ones
        worked += 1
        if worked > buffer_size then
            end_of_buffer = true    -- end of buffer flag
            exit
        end if
    end while
end procedure -- get_worked

procedure iterate()
    integer i,s

    i = worked
    path = buffer[i][1]
    s = buffer[i][2]                -- step

    current = buffer[i][4] - l[path[n]][path[1]] - l[path[s+1]][path[s+2]] +
        l[path[1]][path[s+2]] + l[path[n]][path[s+1]]

    buffer[i][2] -= 1
    if buffer[i][2] = 0 then
        buffer[i][1] = path[2..n] & path[1]
        buffer[i][2] = maxstep
        buffer[i][3] += 1
        if buffer[i][3] > n then
            get_worked()
            maxsize += 1
        end if
    end if
    if worst > current  then
        path[1..1+s] = reverse(path[1..1+s])
        insert()
    elsif (buffer_size < maxsize) and (worst > current)  then
        path[1..1+s] = reverse(path[1..1+s])
        buffer = append(buffer, {path, maxstep, 1, current})
        buffer_size += 1
        worst = current
    end if
end procedure -- iterate

procedure show_iteration()
    select_font(vga8x16)
    reset_font()
    set_attributes(NONE)
    justify(RIGHT)
    set_foreground(15)
    bar(3,553,300,82,17)            -- erase current iteration
    setxy(630,301)
    write(sprint(iteration))
end procedure -- show_iteration

procedure run()
    t = time()
    if length(user_path) then
        user_path = {}
        user_total = 0
    end if
    if iteration >= maxi then
        maxi *= 2
    end if
    while iteration < maxi and not end_of_buffer do
        iterate()
        iteration += 1
        if remainder(iteration, 200) = 0 then
            update_timer()
            if iteration < 1000 or remainder(iteration, 1000) = 0 then
                show_iteration()
            end if
        end if
        if best > current then
            previous = best
            show_previous()
            best = current
            show_best()
            previous_iteration = iteration
            best_path = path
            redraw()
            if self = b[3] then exit end if     -- step
        end if
    end while
    show_iteration()
    if end_of_buffer then
        deactivate(b[2])            -- run
        deactivate(b[3])            -- step
    end if
end procedure -- run

procedure create_node_zones()
    z = repeat({}, n)
    for i=1 to n do
        z[i] = {node[i][1]-4,node[i][2]-4,node[i][1]+4,node[i][2]+4}
    end for
end procedure -- create_node_zones

procedure init()
    maxi = n*n*n                    -- maximum nunber of iterations
    maxsize = 4*n                   -- maximum buffer size
    end_of_buffer = false           -- end of run flag
    maxstep = floor((n+1)/2)
    iteration = 0
    previous_iteration = 0
    erase_user_entry()
    hide_rat()
    bar(3,553,300,82,17)            -- current iteration window
    bar(3,480,260,70,17)            -- previous path window
    bar(3,553,260,82,17)            -- previous iteration window
    show_rat()
    activate(b[2])                  -- run button
    activate(b[3])                  -- step button
    activate(b[6])                  -- redo button
    if n=200 and kind=SCI then
        activate(b[4])              -- save button
    else
        deactivate(b[4])            -- save button
    end if
    if kind = RAND then
        random_nodes()
    elsif kind = POLY then
        poly_nodes()
    elsif kind = GRID then
        grid_nodes()
    else
        sci_nodes()
    end if
    set_buffers()
    create_node_zones()
    redraw()
    if kind=SCI then
        select_font(modern28)
        reset_font()
        justify(CENTRE+BOXED)
        set_foreground(14)
        set_attributes(SHADOW)
        set_shadow_color(0)
        set_extra_char_pitch(1)
        setxy(240,240)
        write("G'day, Arthur !")
    end if
    show_best()
    dt = 0.0                    -- elapsed time
    old_et = 0
    show_time("0:00")
end procedure -- init

procedure help()
    show_reader(re)
end procedure

procedure change_number_of_nodes()  -- select new number of nodes
    n = item(ni, "selected")
    deactivate(b[2])            -- run
    deactivate(b[3])            -- step
    deactivate(b[6])            -- redo
end procedure -- change_number_of_nodes

procedure redo()
    bar(4,7,7,466,466)          -- user field (erase)
    draw_path(12)
    draw_nodes()
    erase_user_entry()
end procedure -- redo

procedure in_rat_zone()
    integer k,len,m,zi

    wait_while(lbp)
    zi = rat_zone(z)
    if zi then
        user_path &= zi
        len = length(user_path)
        hide_rat()
        if len = 1 then
            bar(4,7,7,466,466)      -- red field
            draw_path(12)
            draw_nodes()
        else
            m = user_path[len-1]
            draw_line(15,{{node[m][1],node[m][2]},{node[zi][1],node[zi][2]}})
            ellipse(15, 0, node[m]-2, node[m]+2)
            user_total += l[m][zi]
        end if
        ellipse(14, 0, node[zi]-2, node[zi]+2)
        if len = n then
            k = user_path[1]
            draw_line(15,{{node[zi][1],node[zi][2]},{node[k][1],node[k][2]}})
            user_total += l[k][zi]
        end if
        show_user_total()
        show_rat()
    end if
end procedure -- in_rat_zone

procedure change_kind()
    integer i

    i = item(rbs, "selected")
    kind = find(i, rb)
    deactivate(b[2])            -- run
    deactivate(b[3])            -- step
    deactivate(b[4])            -- save
    deactivate(b[6])            -- redo
end procedure -- change_kind

procedure save()
    integer f

    f = open("tsp.sav", "a")
    printf(f, "{%.1f, %d}", {mu*best, previous_iteration})
    puts(f, "\n")
    print(f, path)
    puts(f, "\n\n")
    close(f)
--    deactivate(b[4])            -- save
end procedure -- save


-- main --------------------------------------------------------------

n = 36                      -- default number of nodes
rb = repeat(0,4)
rb[1] = rbutton("r~a~ndom",485,390,1)
rb[2] = rbutton("~p~olygon",485,408,0)
rb[3] = rbutton("~g~rid",570,390,0)
rb[4] = rbutton("s~c~i am",570,408,0)
for i=1 to 4 do splice(rb[i], routine_id("change_kind")) end for

rbs = rbuttons(rb)           -- rbutton cluster

b = repeat(0,7)

b[1] = button("~n~ew",480,430,50,20)
splice(b[1], routine_id("init"))

b[2] = button("~r~un",532,430,50,20)
splice(b[2], routine_id("run"))

b[3] = button("~s~tep",584,430,50,20)
splice(b[3], routine_id("run"))

b[4] = button("sa~v~e",480,455,37,20)
splice(b[4], routine_id("save"))

b[5] = button("~h~elp",519,455,37,20)
splice(b[5], routine_id("help"))

b[6] = button("re~d~o",558,455,37,20)
splice(b[6], routine_id("redo"))

b[7] = button("~q~uit",597,455,37,20)
splice(b[7], routine_id("quit"))

ni = isel("number of nodes", 530,n,6,200,480,362,48,17)
bind(part(ni, "digits"), "font", vga8x16)
splice(part(ni, "rar"), routine_id("change_number_of_nodes"))
splice(part(ni, "lar"), routine_id("change_number_of_nodes"))

zo = zone(5,5,475,475)      -- right zone, red
arat(zo, routine_id("in_rat_zone"))

re = reader("tsp help", hs, 20,84,440,312)


set_mode(18)
o = palette(12,{63,0,0})    -- make light red a bit darker
widgets = rbs & b & ni & zo
deactivate(b[4])
draw_screen()
kind = RAND                 -- initial node configuration
init()
loop()
set_text_mode()
