--  widgets.e
--  jbabor@paradise.net.nz
--  00-11-22
--  version 2.50

--==================================================================--
--  This version of widgets is a fusion of
--      font.e      version 6.10  (included)
--      rat.e       version 2.50  (included)
--      list.e      version 2.00  (merged)
--      ports.e     Jacques Deschenes (merged: Output only)
--      Pete's smart_print -  merged, mods by Gabriel Boehme & jiri
--      widgets     core code
--      handful of miscellaneous routines
--==================================================================--


without warning
without type_check

include font.e
include rat.e


----------------------------------------------------------------------
--  misc common declaration
----------------------------------------------------------------------

global constant
    false = 0,
    true = 1,
    enter = 13,
    esc = 27,
    space = 32


global function max(sequence s)     -- too simple, but fast!
    atom a
    a = s[1]
    for i=2 to length(s) do
        if a < s[i] then a = s[i] end if
    end for
    return a
end function

global function abs(object x)      -- Carl's
    return x*((x>0)-(x<0))
end function

global function round(object x)
--  simple, universal & fast
--  it rounds mid-points *up*, both negative and positive

    if atom(x) then
        return floor(x+0.5)
    end if
    for i=1 to length(x) do
        x[i] = round(x[i])
    end for
    return x
end function -- round

global procedure put_image(integer x,integer y,sequence image)
--  almost same as display_image()

    for i=1 to length(image) do
        pixel(image[i], {x,y})
        y += 1
    end for
end procedure -- put_image

global procedure xor_image(integer x,integer y,sequence im)
    -- similar to display_image(): 2-d sequence of pixels im is xor-ed with
    -- screen (background) starting at point x,y (left top corner)

    for i=1 to length(im) do
        pixel(xor_bits(get_pixel({x,y}&length(im[i])),im[i]),{x,y})
        y += 1
    end for
end procedure -- xor_image

global procedure merge_image(
    integer x,      -- top left corner, horizontal coordinate
    integer y,      -- top left corner, vertical coordinates
    sequence image, -- 2-d image sequence
    integer c       -- transparent cell color
    )
    --  display 2-d sequence im at point x,y
    --  image cells of color c are considered transparent
    --  with 'on-the-fly' run-length-encoding

    sequence row
    integer f,k,w

    x -= 1
    for i=1 to length(image) do
        row=image[i]
        w = length(row)
        f=0                 -- not in run
        for j=1 to w do
            if row[j]=c then
                if f then
                    pixel(row[k..j-1],{x+k,y})
                    f=0
                end if
            else
                if not f then
                    k=j
                    f=1
                end if
            end if
        end for
        if f then
            pixel(row[k..w],{x+k,y})
        end if
        y+=1
    end for
end procedure -- merge_image

global function replicate(
    sequence pattern,
    integer newlength,  -- required patterned length
    integer offset      -- left edge pattern offset for global tiling
    )
    --  return a sequence generated by pattern repetition
    --  for *local* patterns the offset is *zero*
    --  for *global* patterns the offset is the starting
    --  horizontal location in global coordinates

    if offset then
        offset = remainder(offset,length(pattern))
        newlength += offset
    end if
    while length(pattern) < newlength do
        pattern &= pattern
    end while
    return pattern[1+offset..newlength]
end function -- replicate

global function tile(   -- thanks to C. Taylor for his contribution
    sequence pattern,   -- 2d patern, tile
    integer newwidth,   -- width of tiled area
    integer newheight,  -- height of tiled area
    integer xoffset,    -- left edge for global tiling
    integer yoffset     -- top edge for global tiling
    )
    --  return 2-d sequence (image) of tiled area
    --  for local tiling xoffset and yoffset are zeros, and
    --  the pattern will start in the top left corner of tiled area

    if newheight <= length(pattern) then
        pattern = replicate(pattern, newheight, yoffset)
        for i = 1 to newheight do
            pattern[i] = replicate(pattern[i], newwidth, xoffset)
        end for
        return pattern
    else
        for i = 1 to length(pattern) do
            pattern[i] = replicate(pattern[i], newwidth, xoffset)
        end for
        return replicate(pattern, newheight, yoffset)
    end if
end function -- tile

global procedure bar(
    object c,       -- color or 2-d pattern (tile)
    integer x,      -- left edge
    integer y,      -- top edge
    integer width,
    integer height
    )
    --  draw filled rectangle

    if atom(c) then
        machine_proc(11,{c,1,{{x,y},{x+width-1,y},{x+width-1,y+height-1},{x,y+height-1}}})
    else
        display_image({x,y},tile(c,width,height,x,y))
    end if
end procedure -- bar

global procedure frame(
    object c,       -- if atom(c) then it is the line color,
                    -- else it is frame colors c={{i1,i2,...},{j1,j2...}},
                    -- where i1, i2,... and j1, j2,...  are top left and
                    -- bottom right edges respectively, going from outside
    integer x,      -- left edge
    integer y,      -- top edge
    integer w,      -- width
    integer h       -- height
    )
    if atom(c) then
        machine_proc(11,{c,0,{{x,y},{x+w-1,y},{x+w-1,y+h-1},{x,y+h-1}}})
    else
        for i=1 to length(c[1]) do
            draw_line(c[1][i], {{x+i-1,y+h-i},{x+i-1,y+i-1},{x+w-i-1,y+i-1}})
        end for
        for i=1 to length(c[2]) do
            draw_line(c[2][i], {{x+i,y+h-i},{x+w-i,y+h-i},{x+w-i,y+i-1}})
        end for
    end if
end procedure -- frame


--  Modified extract from ports.e by Jacques Deschenes, Output only:
sequence OutputCode
OutputCode = {
                #50,        -- PUSH EAX
                #52,        -- PUSH EDX
                #BA,0,0,0,0,-- MOV EDX, PORT (port, to be poked in) (3)
                #B0,#00,    -- MOV AL, byte  (byte, to be poked in) (8)
                #EE,        -- OUT DX, AL
                #5A,        -- POP EDX
                #58,        -- POP EAX
                #C3         -- RET
            }
atom OutputAsm, OutputPort, OutputByte

procedure InitPorts()
    OutputAsm = allocate(length(OutputCode))
    if not OutputAsm then
        puts(1,"Error: memory allocation failed in InitPorts()\n")
        abort(1)
    end if
    poke(OutputAsm,OutputCode)
    OutputPort = OutputAsm + 3
    OutputByte = OutputAsm + 8
end procedure -- InitPorts()

procedure Output(integer b, integer port)
    poke4(OutputPort,port)
    poke(OutputByte,b)
    call(OutputAsm)
end procedure -- Output()

InitPorts()


global procedure set_colors(integer c, sequence s)
    --  set contiguous range of colors
    --  c is starting color index
    --  s is sequence of {r,g,b} palette entries
    --  rgb values range 0..63

    Output(c,#3C8)
    for i=1 to length(s) do
        Output(s[i][1],#3C9)
        Output(s[i][2],#3C9)
        Output(s[i][3],#3C9)
    end for
end procedure -- set_colors


-----------------------------------------------------------------
--  list.e  version 2.00  ('localized' & trimmed)
-----------------------------------------------------------------

function index(sequence list, object key)
    --  return *key* index, or zero if key not found

    return find(key, list[1])
end function

function fetch(sequence list, object key)
    --  return list item with indicated key

    integer i

    i = find(key, list[1])
    if i then
        return list[2][i]
    end if
    crash_message("fetch error: key \"" & key & "\" not found...\n")
end function -- fetch

function Fetch(
    object list,        -- list from which value is to be fetched
    sequence keys)      -- sequence of list keys, one for each level
                        -- last one directly associated with fetched
                        -- value
    --  return item from *nested* list : 'deep' fetch

    for i=1 to length(keys) do
        list = fetch(list, keys[i])
    end for
    return list
end function -- Fetch


----------------------------------------------------------------------
-- widgets : initialization & general code
----------------------------------------------------------------------

global constant
    kbd_event = 16,             -- keyboard event
    RGB={52,52,52}              -- default primary color palette (color 8)

global sequence wid,pile,widgets
global integer exitf,pressed_key,hsize,vsize,menuf,pilef,pilep,screen,self
sequence clicked, keyed, rgb
integer emask                   -- event mask

global procedure set_default_colors(
    sequence mix -- mix = {red,green,blue} primary face color mix
    )
    -- set graduated scale of standard colors (0,3,7,8,11,15)

    object p
    p = {
        {0,0,0},            -- black, text
        {0,0,42},           -- blue
        {0,42,0},           -- green
        floor(mix*0.50),    -- shadows
        {42,0,0},           -- red
        {42,0,42},          -- magenta
        {42,21,0},          -- brown
        floor(mix*0.85),    -- face
        mix,                -- light face ('paper'color)
        {21,21,63},         -- light blue
        {21,63,21},         -- light green
        floor((mix+64)/2),  -- highlights
        {63,21,21},         -- light red
        {63,21,63},         -- light magenta
        {63,63,21},         -- yellow
        {63,63,63}}         -- white
    set_colors(0, p)
end procedure -- set_default_colors

global procedure set_mode(integer mode)
    --  invoke desired graphics mode and set standard colors

    object vc

    use_vesa(1)     -- just in case...
    if graphics_mode(mode) then
        puts(1,"Requested graphics mode not available!\n")
    end if
    set_default_colors(rgb)
    vc = video_config()
    hsize = vc[VC_XPIXELS]
    vsize = vc[VC_YPIXELS]
    init_rat()
end procedure -- set_mode

global procedure set_text_mode()
    if graphics_mode(-1) then end if
end procedure -- set_text_mode

global procedure zap(integer w)
    -- zap widget and all its components to recover some memory

    sequence c,s

    s = wid[w]
    if index(s, "parts") then
        c = fetch(s, "parts")
        for i=1 to length(c[1]) do
            zap(c[2][i])
        end for
    end if
    wid[w] = {}
end procedure -- zap

global procedure shift(integer w, integer dx, integer dy)
    -- translate widget and all its components by indicated
    -- number of pixels in both directions

    sequence c,s
    integer j,k,m

    s = wid[w]
    if index(s, "parts") then
        c = fetch(s, "parts")
        for i=1 to length(c[1]) do
            shift(c[2][i], dx, dy)
        end for
    end if
    j = index(s, "x")
    if j then
        s[2][j] += dx
    end if
    k = index(s, "y")
    if k then
        s[2][k] += dy
    end if
    m = index(s,"box")
    if m then
        s[2][m][1] += dx
        s[2][m][2] += dy
    end if
    if j or k or m then
        wid[w] = s
    end if
end procedure -- shift


----------------------------------------------------------------------
--  items (properties & methods) and components
----------------------------------------------------------------------

global function item(       -- fetch named item of widget, property
    integer wi,             -- widget (index)
    object name             -- item name, key
    )
    return fetch(wid[wi], name)
end function -- item

global procedure bind(
    integer wi,             -- widget (index)
    object name,            -- item name, key
    object val              -- new value
    )
    -- replace current item value with new one

    wid[wi][2][find(name, wid[wi][1])] = val
end procedure -- bind

global procedure make(
    integer wi,             -- widget (index)
    object name,            -- item name, key
    object val              -- new value
    )
    -- add new item to list if it does not exist,
    -- otherwise just reset its value

    integer i

    i = index(wid[wi], name)
    if i then
        wid[wi][2][i] = val
    else
        wid[wi][1] = append(wid[wi][1], name)
        wid[wi][2] = append(wid[wi][2], val)
    end if
end procedure -- make

global function part(   
    integer wi,         -- widget (index)
    object name         -- part name
    )
    -- return wid index of named component

    return Fetch(wid[wi], {"parts", name})
end function -- part

global procedure affix( 
    integer wm,         -- 'boss' widget (index)
    object name,        -- part name, key
    integer wi          -- (new) part stack index
    )
    -- add new part

    integer i

    i = index(wid[wm], "parts")
    if i then           -- parts already exit, append them
        wid[wm][2][i][1] = append(wid[wm][2][i][1], name)
        wid[wm][2][i][2] = append(wid[wm][2][i][2], wi)
    else                -- parts do not exit, generate them
        make(wm, "parts", {{name},{wi}})
    end if
    make(wi, "main", wm)
end procedure -- affix

global function main(integer wi)
    -- return wid index of 'main' widget

    integer i

    i = find("main", wid[wi][1])
    if i then
        return wid[wi][2][i]
    end if
    return 0
end function


----------------------------------------------------------------------
-- 'generic' methods
----------------------------------------------------------------------

global procedure show(object wi)
    if atom(wi) then
        call_proc(item(wi, "show"), {wi})
    else
        for i=1 to length(wi) do
            call_proc(item(wi[i], "show"), {wi[i]})
        end for
    end if
end procedure -- show

global procedure hide(object wi)
    if atom(wi) then
        call_proc(item(wi, "hide"), {wi})
    else
        for i=1 to length(wi) do
            call_proc(item(wi[i], "hide"), {wi[i]})
        end for
    end if
end procedure -- hide

global procedure on_rat()
    sequence s

    s = item(self, "rat")
    for i=1 to length(s) do
        call_proc(s[i], {})
    end for
end procedure -- on_rat

global procedure on_key()
    sequence s

    s = item(self, "kbd")
    for i=1 to length(s) do
        call_proc(s[i], {})
    end for
end procedure -- on_key

global procedure arat(integer wi, integer rid)
    -- append routine_id rid of a method to widget's "rat"
    -- sequence of routine ids

    sequence s

    s = item(wi, "rat")
    bind(wi, "rat", s & rid)
end procedure -- arat

global procedure akbd(integer wi, integer rid)
    -- append routine_id rid of a method to widget's 'kbd'
    -- sequence of routine ids

    sequence s

    s = item(wi, "kbd")
    bind(wi, "kbd", s & rid)
end procedure -- akbd

global procedure splice(integer wi, integer rid)
    -- append routine_id rid of a method to both 'rat' and 'kbd
    -- sequences of routine ids

    sequence s

    s = item(wi, "rat")
    bind(wi, "rat", s & rid)
    s = item(wi, "kbd")
    bind(wi, "kbd", s & rid)
end procedure -- splice

----------------------------------------------------------------------
-- zones & tracking
----------------------------------------------------------------------

procedure Track(integer w)
    -- get all rat zones traps and key traps in widget w

    sequence k,p,s,u
    integer e,i,l,x,y

    p = wid[w]
    i = find("event",p[1])          -- index
    if i then
        e = p[2][i]
        if and_bits(e, kbd_event) then
            k = fetch(p, "key")
            l = length(k)
            if l then
                for j=1 to l do
                    keyed = append(keyed, {w,k[j]})
                end for
                emask = or_bits(emask, kbd_event) -- events to be tracked
            end if
        end if
        e = and_bits(e, rat_events)
        if e then
            u = item(w, "box")
            x = u[1]
            y = u[2]
            clicked = append(clicked, {w , e, x, y, x+u[3]-1, y+u[4]-1})
            emask = or_bits(emask,e)    -- events to be tracked
        end if
    end if
    i = find("parts", p[1])             -- index
    if i then
        s = p[2][i][2]  -- part sequence
        for j=1 to length(s) do
            Track(s[j])
        end for
    end if
end procedure -- Track

global procedure track(object ws)
    -- collect all traps in all widgets in sequence ws, recursively,
    -- into easily 'scannable' structures:
    --  "clicked" : {{widget_index, event, x1, y1, x2, y2}, {...
    --  "keyed" : {{widget_index, key}, {...,
    --  where key is ascii code

    emask = 0       -- events to be tracked
    clicked = {}
    keyed = {}
    for i=1 to length(ws) do
        Track(ws[i])
    end for
end procedure -- track

global function get_event(integer mask)
    -- update rat and wait for specified event
    -- return event code

    integer e

    while 1 do
        rat()
        e = and_bits(rat_event, mask)       -- check rat first
        if e then
            return e
        end if
        pressed_key = get_key()
        if pressed_key != -1 then
            if and_bits(mask, kbd_event) or pressed_key = esc then
                return kbd_event
            end if
        end if
    end while
end function -- get_event

global procedure collapse_menu()
    integer l,wi

    l = length(pile)
    while l > pilep do
        wi = pile[l]        -- submenu on top of pile
        hide(wi)
        l -= 1
        pile = pile[1..l]
    end while
    menuf = false
end procedure -- collapse_menu

global procedure loop()
    -- track widgets, including their parts and
    -- respond to *all* events as they occur

    sequence zi
    integer event,flag,len

    exitf = false
    if screen then
        widgets &= screen
    end if
    show_rat()
    while not exitf do
        len = length(pile)
        if len then             -- window or submenu in use
            track({pile[len]})  -- track the topmost item
        else                    -- no more menus nor windows
            track(widgets)
        end if
        pilef = false
        while not exitf and not pilef do
            event = get_event(emask)
            if event = kbd_event then
                flag = false
                for i=1 to length(keyed) do
                    if keyed[i][2] = pressed_key then
                        self = keyed[i][1]
                        on_key()
                        flag = true
                        exit                -- exit for loop
                    end if
                end for
                if flag = false and pressed_key = esc then
                    exitf = true
                end if
            end if
            if and_bits(event, lbp)  then
                flag = false
                for i=1 to length(clicked) do
                    zi = clicked[i]
                    if and_bits(zi[2], lbp) and
                        rx>=zi[3] and ry>=zi[4] and rx<=zi[5]
                        and ry<=zi[6] then
                            flag = true
                            self = zi[1]
                            on_rat()
                            exit            -- exit for loop
                    end if
                end for
                if not flag and menuf then  -- click outside: collapse menus
                    collapse_menu()
                end if
            end if
        end while
    end while
end procedure -- loop


----------------------------------------------------------------------
-- text : built-in text widget
----------------------------------------------------------------------

include sysb13.e    -- font

global integer default_font
integer m

default_font = sysb13

global function text(
    integer x,
    integer y,
    sequence s
    )
    wid = append(wid, {{"x","y","string","show","font","fc",
        "hc","sc","paper","attrib","just","ecp","elp","hso","vso"},
        {x,y,s,m,default_font,BLACK,BRIGHT_RED,3,7, HILITE, LEFT, 0, 1, 1, 1}})
    return length(wid)
end function -- text

global procedure retext(        -- new word: REset TEXT environment
    integer wi                  -- text widget
    )
    sequence p

    p = wid[wi][2]              -- values sequence of text list
    select_font(p[5])
    set_foreground(p[6])
    set_hilite_color(p[7])
    set_shadow_color(p[8])
    set_paper_color(p[9])
    set_attributes(p[10])
    justify(p[11])
    set_extra_char_pitch(p[12])
    set_extra_line_pitch(p[13])
    set_shadow_offsets(p[14], p[15])
end procedure -- retext

global procedure show_text(
    integer wi                  -- text widget
    )
    sequence p

    p = wid[wi][2]              -- values sequence of text list
    retext(wi)
    setxy(p[1], p[2])
    hide_rat()
    write(p[3])
    show_rat()
end procedure -- show_text
m = routine_id("show_text")


----------------------------------------------------------------------
-- activate / deactivate : auxiliary global routines
----------------------------------------------------------------------

global constant DIMMED = {{3,7},{7,3}}

global procedure activate(integer wi    -- reactivate this widget
    )
    integer wt

    if not find(wi, widgets) then       -- start tracking it again
        widgets &= wi
        wt = part(wi, "label")          -- text
        bind(wt, "fc", BLACK)           -- foreground color
        bind(wt, "hc", BRIGHT_RED)      -- hilite color
        show(wi)
        track(widgets)
    end if
end procedure -- activate

global procedure deactivate(integer wi  -- deactivate this widget
    )
    integer i,wt

    i = find(wi, widgets)
    if i then                           -- remove it from tracked bunch
        widgets = widgets[1..i-1] & widgets[i+1..length(widgets)]
        wt = part(wi, "label")          -- text
        bind(wt, "fc", DIMMED)
        bind(wt, "hc", DIMMED)
        show(wi)
        track(widgets)
    end if
end procedure -- deactivate


----------------------------------------------------------------------
-- initialize widgets core
----------------------------------------------------------------------

rgb = RGB       -- default: RGB = {52,52,52}
screen = 0      -- default, index of screen closer widget
widgets = {}    -- trackable widgets (default)
menuf = 0       -- no active menu
wid = {}        -- stack of widgets
pile = {}       -- windows & menus pile
pilep = 0       -- pile pointer : points to top of pile just before
                -- current menu or menus were added
--tick_rate(200)


----------------------------------------------------------------------
--  Pete's smart_print -  mods by Gabriel Boehme & me
----------------------------------------------------------------------

function smart_print_string(sequence s)
--  faster, I think, as a function

    object x

    for i = 1 to length(s) do
        x = s[i]
        if not integer(x) or x < ' ' or x > 126 or x = '"' then
            return 0
        end if
    end for
    return 1
end function

global procedure sp(object o)  -- smart print
    if atom(o) then
        printf(1, "%g", {o})
    elsif length(o) = 0 then
        puts(1, "{}")
    elsif smart_print_string(o) then
        printf(1, "\"%s\"", {o})
    else
        puts(1, "{")
        sp(o[1])
        for i = 2 to length(o) do
            puts(1, ",")
            sp(o[i])
        end for
        puts(1, "}")
    end if
end procedure -- sp

global procedure spn(object o) -- smart print & newline
    sp(o)
    puts(1, "\n")
end procedure -- spn


----------------------------------------------------------------------
-- debugging
----------------------------------------------------------------------

procedure show_parts(sequence s)
    integer len

    len = length(s[1])
    puts(1, "(")
    for i=1 to len do
        printf(1, "%s:%d", {s[1][i], s[2][i]})
        if i=len then
            puts(1, ")")
        else
            puts(1, ", ")
        end if
    end for
end procedure

global procedure pw(object w)
    --  formatted print of widget sequence

    sequence wi
    integer len

    if atom(w) then w = {w} end if
    for i=1 to length(w) do
        text_color(BRIGHT_WHITE)
        sp(w[i])
        text_color(WHITE)
        puts(1, ": (")
        wi = wid[w[i]]
        len = length(wi[1])
        for j=1 to len do
            printf(1, "%s:", {wi[1][j]})
            if equal(wi[1][j], "parts") then
                show_parts(wi[2][j])
            else
                sp(wi[2][j])
            end if
            if j=len then
                puts(1, ")\n")
            else
                puts(1, ", ")
            end if
        end for
    end for
end procedure -- pw
